/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 2002 The Apache Software Foundation.  All rights 
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:  
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "WSIF" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written 
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 2001, 2002, International
 * Business Machines, Inc., http://www.apache.org.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

package org.apache.wsif.base;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.apache.wsif.WSIFException;
import org.apache.wsif.WSIFConstants;
import org.apache.wsif.WSIFMessage;
import org.apache.wsif.WSIFPort;
import org.apache.wsif.logging.Trc;

/**
 * A DefaultWSIFPort is a default implementation of WSIFPort
 * all methods are implemented except execute*.
 *
 * @author Paul Fremantle
 * @author Alekander Slominski
 * @author Matthew J. Duftler
 * @author Sanjiva Weerawarana
 * @author Nirmal Mukhi
 */
public abstract class WSIFDefaultPort implements WSIFPort {
	private static final long serialVersionUID = 1L;
	
	private WSIFMessage context;

    public void close() throws WSIFException {
        Trc.entry(this);
        Trc.exit();
    }

    public void finalize() throws Throwable {
  	    Trc.entry(this);
  	    try {
  	        close();
  	    } catch (WSIFException ex) {
            Trc.ignoredException( ex );
  	    }
        super.finalize();
  	    Trc.exit();
    }
  
    /**
     * Utility method to return key suitable for hash table.
     */
    protected String getKey(String name, String inputName, String outputName) {
        Trc.entry(this, name, inputName, outputName);
        String s =
            name
                + (inputName != null ? ":" + inputName : "")
                + (outputName != null ? ":" + outputName : "");
        Trc.exit(s);
        return s;
    }

    /**
     * Utility method to retrieve extensibility element from list
     * checks also that it is exactly one extensibility element.
     */
    protected Object getExtElem(Object ctx, Class extType, List extElems)
        throws WSIFException {
        Trc.entry(this, ctx, extType, extElems);

        Object found = null;
        if (extElems != null) {
            for (Iterator i = extElems.iterator(); i.hasNext();) {
                // if so return new
                Object o = i.next();
                if (extType.isAssignableFrom(o.getClass())) {
                    if (found != null) {
                        throw new WSIFException(
                            "duplicated extensibility element "
                                + extType.getClass().getName()
                                + " in "
                                + ctx);
                    }
                    found = o;
                }
            }
        }
        Trc.exit(found);
        return found;
    }

    /**
     * Utility method to retrieve multiple extensibility elements from a list.
     */
    protected List getExtElems(Object ctx, Class extType, List extElems)
        throws WSIFException {
        Trc.entry(this, ctx, extType, extElems);
        List found = new ArrayList();
        if (extElems != null)
            for (Iterator i = extElems.iterator(); i.hasNext();) {
                Object o = i.next();
                if (extType.isAssignableFrom(o.getClass()))
                    found.add(o);
            }
        if (found.size() == 0)
            return null;
        Trc.exit(found);
        return found;
    }

    /**
     * Tests if this port supports synchronous calls to operations.
     * 
     * @return true   by default WSIFPorts do support synchronous calls
     */
    public boolean supportsSync() {
        Trc.entry(this);
        Trc.exit(true);
        return true;
    }

    /**
     * Tests if this port supports asynchronous calls to operations.
     * 
     * @return false   by default ports do not support asynchronous calls
     */
    public boolean supportsAsync() {
        Trc.entry(this);
        Trc.exit(false);
        return false;
    }

    /**
     * Gets the context information for this WSIFPort.
     * @return context
     */
    public WSIFMessage getContext() throws WSIFException {
        Trc.entry(this);
    	WSIFMessage contextCopy;
    	if (this.context == null) {
    		// really this should call getContext on the WSIFService but
    		// theres no reference to that so WSIFService must call setContext
    		// on any WSIFPorts it creates.
    		contextCopy = new WSIFDefaultMessage();
    	} else {
		    try {
			    contextCopy = (WSIFMessage) this.context.clone();
		    } catch (CloneNotSupportedException e) {
			    throw new WSIFException(
			        "CloneNotSupportedException cloning context", e);
		    }
    	}
        Trc.exit(contextCopy);
    	return contextCopy;
    }

    /**
     * Sets the context information for this WSIFPort.
     * @param WSIFMessage the new context information
     */
    public void setContext(WSIFMessage context) {
        Trc.entry(this, context);
        if (context == null) {
        	throw new IllegalArgumentException("context must not be null");
        }
        this.context = context;
        Trc.exit(null);
    }

}