/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 2002 The Apache Software Foundation.  All rights 
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:  
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "WSIF" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written 
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 2001, 2002, International
 * Business Machines, Inc., http://www.apache.org.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

package org.apache.wsif.base;

import java.util.Hashtable;
import java.util.Map;

import javax.wsdl.Definition;
import javax.wsdl.PortType;
import javax.wsdl.Service;
import javax.xml.namespace.QName;

import org.apache.wsif.WSIFException;
import org.apache.wsif.WSIFService;
import org.apache.wsif.WSIFServiceFactory;
import org.apache.wsif.logging.Trc;

/**
 * Factory class used to create instances of WSIFService
 * 
 * @author Mark Whitlock
 * @author Owen Burroughs <owenb@apache.org>
 */
public class WSIFServiceFactoryImpl extends WSIFServiceFactory {

    private boolean useCache = false;
    private Map cache = new Hashtable();

    /**
     * Create a WSIFService from WSDL document URL.
     * <br> If serviceName or serviceNS is null,
     *   then WSDL document must have exactly one service in it.
     * <br> If portTypeName or portTypeNS is null,
     *   then WSDL document must have exactly one portType in it
     *   and all ports of the selected service must
     *    implement the same portType.
     * @param wsdlLoc The URL for the wsdl's location
     * @param serviceNS The namespace of the service
     * @param serviceName The name of the service
     * @param portTypeNS The namespace of the port type
     * @param portTypeName The name of the port type
     * @return The service
     * @exception A WSIFException if an error occurs when creating the service
     */
    public WSIFService getService(
        String wsdlLoc,
        String serviceNS,
        String serviceName,
        String portTypeNS,
        String portTypeName)
        throws WSIFException {
        Trc.entry(
            this,
            wsdlLoc,
            serviceNS,
            serviceName,
            portTypeNS,
            portTypeName);

        String key = "";
        if (useCache) {
            key =
                genCacheKey(
                    wsdlLoc,
                    serviceNS,
                    serviceName,
                    portTypeNS,
                    portTypeName);
            WSIFServiceImpl cachedWSI = (WSIFServiceImpl) cache.get(key);
            if (cachedWSI != null) {
                WSIFServiceImpl wsi = new WSIFServiceImpl(cachedWSI);
                Trc.exit(wsi);
                return wsi;
            }
        }

        WSIFServiceImpl wsi =
            new WSIFServiceImpl(
                wsdlLoc,
                serviceNS,
                serviceName,
                portTypeNS,
                portTypeName);

        if (useCache && !key.equals("")) {
            cache.put(key, wsi);
        }

        Trc.exit(wsi);
        return wsi;
    }

    /**
     * Create a WSIF service instance from WSDL document URL
     * using a ClassLoader to find local resources.
     * <br> If serviceName or serviceNS is null,
     *   then WSDL document must have exactly one service in it.
     * <br> If portTypeName or portTypeNS is null,
     *   then WSDL document must have exactly one portType in it
     *   and all ports of the selected service must
     *    implement the same portType.
     * @param wsdlLoc The URL for the wsdl's location
     * @param cl A ClassLoader to use in locating the wsdl
     * @param serviceNS The namespace of the service
     * @param serviceName The name of the service
     * @param portTypeNS The namespace of the port type
     * @param portTypeName The name of the port type
     * @return The service
     * @exception A WSIFException if an error occurs when creating the service
     */
    public WSIFService getService(
        String wsdlLoc,
        ClassLoader cl,
        String serviceNS,
        String serviceName,
        String portTypeNS,
        String portTypeName)
        throws WSIFException {
        Trc.entry(
            this,
            wsdlLoc,
            cl,
            serviceNS,
            serviceName,
            portTypeNS,
            portTypeName);

        String key = "";
        if (useCache) {
            key =
                genCacheKey(
                    wsdlLoc,
                    serviceNS,
                    serviceName,
                    portTypeNS,
                    portTypeName);
            WSIFServiceImpl cachedWSI = (WSIFServiceImpl) cache.get(key);
            if (cachedWSI != null) {
                WSIFServiceImpl wsi = new WSIFServiceImpl(cachedWSI);
                Trc.exit(wsi);
                return wsi;
            }
        }

        WSIFServiceImpl wsi =
            new WSIFServiceImpl(
                wsdlLoc,
                cl,
                serviceNS,
                serviceName,
                portTypeNS,
                portTypeName);

        if (useCache && !key.equals("")) {
            cache.put(key, wsi);
        }

        Trc.exit(wsi);
        return wsi;
    }

    /**
     * Returns a new WSIFService.
     * @param def The Definition object representing the wsdl
     * @return The service
     * @exception A WSIFException if an error occurs when creating the service
     */
    public WSIFService getService(Definition def) throws WSIFException {
        Trc.entryExpandWsdl(this, new Object[] { def });
        String key = "";
        if (useCache) {
            key = genCacheKey(def, null, null);
            WSIFServiceImpl cachedWSI = (WSIFServiceImpl) cache.get(key);
            if (cachedWSI != null) {
                WSIFServiceImpl wsi = new WSIFServiceImpl(cachedWSI);
                Trc.exit(wsi);
                return wsi;
            }
        }

        WSIFServiceImpl wsi = new WSIFServiceImpl(def);
        if (useCache && !key.equals("")) {
            cache.put(key, wsi);
        }

        Trc.exit(wsi);
        return wsi;
    }

    /**
     * Returns a new WSIFService.
     * @param def The Definition object representing the wsdl
     * @param service The Service object representing the service to use
     * @return The service
     * @exception A WSIFException if an error occurs when creating the service
     */
    public WSIFService getService(Definition def, Service service)
        throws WSIFException {
        Trc.entryExpandWsdl(this, new Object[] { def, service });
        String key = "";
        if (useCache) {
            key = genCacheKey(def, service, null);
            WSIFServiceImpl cachedWSI = (WSIFServiceImpl) cache.get(key);
            if (cachedWSI != null) {
                WSIFServiceImpl wsi = new WSIFServiceImpl(cachedWSI);
                Trc.exit(wsi);
                return wsi;
            }
        }

        WSIFServiceImpl wsi = new WSIFServiceImpl(def, service);
        if (useCache && !key.equals("")) {
            cache.put(key, wsi);
        }
        Trc.exit(wsi);
        return wsi;
    }

    /**
     * Returns a new WSIFService.
     * @param def The Definition object representing the wsdl
     * @param service The Service object representing the service to use
     * @param portType The PortType object representing the port type to use
     * @return The service
     * @exception A WSIFException if an error occurs when creating the service
     */
    public WSIFService getService(
        Definition def,
        Service service,
        PortType portType)
        throws WSIFException {
        Trc.entryExpandWsdl(this, new Object[]{def, service, portType});
        String key = "";
        if (useCache) {
            key = genCacheKey(def, service, portType);
            WSIFServiceImpl cachedWSI = (WSIFServiceImpl) cache.get(key);
            if (cachedWSI != null) {
                WSIFServiceImpl wsi = new WSIFServiceImpl(cachedWSI);
                Trc.exit(wsi);
                return wsi;
            }
        }

        WSIFServiceImpl wsi = new WSIFServiceImpl(def, service, portType);
        if (useCache && !key.equals("")) {
            cache.put(key, wsi);
        }

        Trc.exit(wsi);
        return wsi;
    }

    /**
     * Returns a new WSIFService.
     * @param def The Definition object representing the wsdl
     * @param serviceNS The namespace of the service
     * @param serviceName The name of the service
     * @param portTypeNS The namespace of the port type
     * @param portTypeName The name of the port type
     * @return The service
     * @exception A WSIFException if an error occurs when creating the service
     */
    public WSIFService getService(
        Definition def,
        String serviceNS,
        String serviceName,
        String portTypeNS,
        String portTypeName)
        throws WSIFException {
        Trc.entryExpandWsdl(
            this,
            new Object[] {
                def,
                serviceNS,
                serviceName,
                portTypeNS,
                portTypeName });

        String key = "";
        if (useCache) {
            key =
                genCacheKey(
                    def,
                    serviceNS,
                    serviceName,
                    portTypeNS,
                    portTypeName);
            WSIFServiceImpl cachedWSI = (WSIFServiceImpl) cache.get(key);
            if (cachedWSI != null) {
                WSIFServiceImpl wsi = new WSIFServiceImpl(cachedWSI);
                Trc.exit(wsi);
                return wsi;
            }
        }

        WSIFServiceImpl wsi =
            new WSIFServiceImpl(
                def,
                serviceNS,
                serviceName,
                portTypeNS,
                portTypeName);
        if (useCache && !key.equals("")) {
            cache.put(key, wsi);
        }

        Trc.exit(wsi);
        return wsi;
    }

    /**
     * Set caching of servies on/off. The default is off. If caching is on then
     * a call to getService will first check if a service matching the parameters
     * specified has already been created and if so a reference to that instance
     * of WSIFService is returned.
     * @param on Flag to indicate whether or not caching of services should be used
     */
    public void cachingOn(boolean on) {
    	Trc.entry(this,on);
        useCache = on;
        Trc.exit();
    }

    private String genCacheKey(
        Definition def,
        Service service,
        PortType portType) {
        Trc.entry(this, def, service, portType);

        String db =
            (def != null && def.getDocumentBaseURI() != null)
                ? def.getDocumentBaseURI()
                : "null";
        QName serviceName = (service != null) ? service.getQName() : null;
        String sn = (serviceName != null) ? serviceName.toString() : "null";
        QName portTypeName = (portType != null) ? portType.getQName() : null;
        String ptn = (portTypeName != null) ? portTypeName.toString() : "null";
        StringBuffer key = new StringBuffer();
        key.append("D=");
        key.append(db);
        key.append("S=");
        key.append(sn);
        key.append("P=");
        key.append(ptn);

        String ret = key.toString();
        // If no distinguishable information is available then don't add to cache
        if (ret.equals("D=nullS=nullP=null")) 
            ret = "";
        Trc.exit(ret);
        return ret;
    }

    private String genCacheKey(
        String wsdlLoc,
        String serviceNS,
        String serviceName,
        String portTypeNS,
        String portTypeName) {
        Trc.entry(
            this,
            wsdlLoc,
            serviceNS,
            serviceName,
            portTypeNS,
            portTypeName);

        StringBuffer key = new StringBuffer();
        if (wsdlLoc == null)
            wsdlLoc = "";
        if (serviceNS == null)
            serviceNS = "";
        if (serviceName == null)
            serviceName = "";
        if (portTypeNS == null)
            portTypeNS = "";
        if (portTypeName == null)
            portTypeName = "";
        key.append("W=");
        key.append(wsdlLoc);
        key.append("SN=");
        key.append(serviceNS);
        key.append("SS=");
        key.append(serviceName);
        key.append("PN=");
        key.append(portTypeNS);
        key.append("PS=");
        key.append(portTypeName);

        String ret = key.toString();
        // If no distinguishable information is available then don't add to cache
        if (ret.equals("W=SN=SS=PN=PS="))
            ret = "";
        Trc.exit(ret);
        return ret;
    }

    private String genCacheKey(
        Definition def,
        String serviceNS,
        String serviceName,
        String portTypeNS,
        String portTypeName) {
        Trc.entry(this, def, serviceNS, serviceName, portTypeNS, portTypeName);

        StringBuffer key = new StringBuffer();
        String db =
            (def != null && def.getDocumentBaseURI() != null)
                ? def.getDocumentBaseURI()
                : "null";
        if (serviceNS == null)
            serviceNS = "";
        if (serviceName == null)
            serviceName = "";
        if (portTypeNS == null)
            portTypeNS = "";
        if (portTypeName == null)
            portTypeName = "";
        key.append("D=");
        key.append(db);
        key.append("SN=");
        key.append(serviceNS);
        key.append("SS=");
        key.append(serviceName);
        key.append("PN=");
        key.append(portTypeNS);
        key.append("PS=");
        key.append(portTypeName);

        String ret = key.toString();
        // If no distinguishable information is available then don't add to cache
        if (ret.equals("D=nullSN=SS=PN=PS="))
            ret = "";
        Trc.exit(ret);
        return ret;
    }
}
