/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 2002 The Apache Software Foundation.  All rights 
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:  
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "WSIF" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written 
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 2001, 2002, International
 * Business Machines, Inc., http://www.apache.org.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */

package org.apache.wsif.util;

import java.io.InputStream;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.Properties;

import org.apache.wsif.WSIFConstants;
import org.apache.wsif.logging.Trc;

/**
 * The WSIFProperties class contains various static methods to read
 * values from the wsif.properties file. Its main purpose is to prevent the
 * properties file from being read in from disk each time a property is checked.
 * 
 * @author Ant Elder <antelder@apache.org>
 */
public class WSIFProperties {

    private static Properties properties;

    /**
     * Reads a property from the wsif.properties file.
     * 
     * @param property   the property to read
     * @return the value of ther property or null if the property is not defined   
     */
    public static String getProperty(String property) {
        Trc.entry(null, property);
        if (properties == null) {
            properties =
                (Properties) AccessController.doPrivileged(new PrivilegedAction() {
                public Object run() {
                    InputStream in = (Thread.currentThread().getContextClassLoader())
                       .getResourceAsStream(WSIFConstants.WSIF_PROPERTIES);

                    Properties p2 = new Properties();
                    try {
                        p2.load(in);
                    } catch (Exception ignored) {
			        	Trc.exception(ignored);
                        return null;
                    }
                    return p2;
                }
            });
        }

        if (properties == null) {
            Trc.exit(null);
            return null;
        }

        String s = properties.getProperty(property);
        Trc.exit(s);
        return s;
    }

    /**
     * Reads the async request timeout value from the wsif.properties file. This 
     * property defines how long WSIF will keep a WSIFOperation stored in the
     * correlation service waiting for the reply to a WSIFOperation 
     * executeRequestResponseAsync method call.
     * 
     * @return the async request timeout value in milliseconds. If the property
     *         is invalid or not defined in the properties file then a value of
     *         zero is returned.
     */
    public static long getAsyncTimeout() {
        Trc.entry(null);
        long t;
        try {
            t = Long.parseLong(getProperty(WSIFConstants.WSIF_PROP_ASYNC_TIMEOUT));
            if (t < 0) {
                t = 0;
            } else {
                t = t * 1000; // convert to milliseconds
            }
        } catch (NumberFormatException e) {
        	Trc.exception(e);
            t = 0;
        }
        Trc.exit(new Long(t));
        return t;
    }

    /**
     * Reads the synchronous request timeout from the wsif.properties file. This defines
     * how long (in milliseconds) WSIF will wait for a synchronous response. 
     * Default of 0 means forever.
     */
    public static long getSyncTimeout() {
        Trc.entry(null);
        long t;
        try {
            t = Long.parseLong(getProperty(WSIFConstants.WSIF_PROP_SYNC_TIMEOUT));
            if (t < 0) {
                t = 0;
            }
        } catch (NumberFormatException e) {
        	Trc.exception(e);
            t = 0;
        }
        Trc.exit(new Long(t));
        return t;
    }
}