/*
 * The Apache Software License, Version 1.1
 *
 *
 * Copyright (c) 2002 The Apache Software Foundation.  All rights 
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:  
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "WSIF" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written 
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation and was
 * originally based on software copyright (c) 2001, 2002, International
 * Business Machines, Inc., http://www.apache.org.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */
 
package org.apache.wsif.wsdl;
 
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.net.URL;

import org.apache.wsif.logging.Trc;

import com.ibm.wsdl.util.StringUtils;

/**
 * Implementation of javax.wsdl.xml.WSDLLocator. This class can be used to
 * locate a wsdl document and its imports using a ClassLoader. This is useful
 * when the wsdl is located in a jar/zip file.
 * 
 * @author Owen Burroughs <owenb@apache.org>
 */ 
public class WSIFWSDLLocatorImpl implements javax.wsdl.xml.WSDLLocator { 
 
    Reader baseReader = null; 
    Reader importReader = null; 
    String contextURI = null; 
    String wsdlLocation = null; 
    String documentBase = null; 
    String importBase = null; 
    ClassLoader loader = null; 

	/**
	 * Create an instance of WSIFWSDLLocatorImpl.
	 * @param ctxt The context uri for the wsdl location
	 * @param wsdlURI The uri for the base wsdl document
	 * @param cl A ClassLoader to use in locating the base wsdl document and imports
	 */ 
    public WSIFWSDLLocatorImpl(String ctxt, String wsdlURI, ClassLoader cl) { 
    	Trc.entry(this,ctxt,wsdlURI,cl);
        contextURI = ctxt; 
        wsdlLocation = wsdlURI; 
        loader = cl; 
        Trc.exit();
    }

	/**
	 * Create an instance of WSIFWSDLLocatorImpl.
	 * @param docBase The uri for the base wsdl document
	 * @param reader A reader "directed at" the base wsdl document
	 * @param cl A ClassLoader to use in locating the base wsdl document and imports
	 */     
    public WSIFWSDLLocatorImpl(String docBase, Reader reader, ClassLoader cl) { 
    	Trc.entry(this,docBase,cl);
        documentBase = docBase; 
        baseReader = reader; 
        loader = cl; 
        Trc.exit();
    } 

	/**
	 * Get a reader for the base wsdl document. Returns null if the document
	 * cannot be located.
	 * @return The reader or null if the import cannot be resolved
	 */  
    public Reader getBaseReader() { 
    	Trc.entry(this);
        if (baseReader == null) { 
            try { 
                URL url = null; 
                URL contextURL = 
                    (contextURI != null) ? StringUtils.getURL(null, contextURI) : null; 
                if (loader != null) { 
                    InputStream in = null; 
                    try { 
                        if (contextURL != null) 
                            url = new URL(contextURL, wsdlLocation); 
                        else { 
                            if (wsdlLocation.indexOf(":") == -1) 
                                url = new URL("file", null, wsdlLocation); 
                            else 
                                url = new URL(wsdlLocation); 
                        } 
                        String wsdlRelativeLocation = url.getPath(); 
                        if (wsdlRelativeLocation.startsWith("/")) 
                            wsdlRelativeLocation = wsdlRelativeLocation.substring(1); 
                        in = loader.getResourceAsStream(wsdlRelativeLocation); 
                        baseReader = new InputStreamReader(in); 
                    } catch (Exception exc) { 
			        	Trc.ignoredException(exc);
                    } 
                } 
                if (baseReader == null) { 
                    url = StringUtils.getURL(contextURL, wsdlLocation); 
                    baseReader = StringUtils.getContentAsReader(url); 
                } 
                if (url != null) 
                    documentBase = url.toString(); 
            } catch (Exception e) {
            	Trc.exception(e);
            	documentBase = wsdlLocation; 
            } 
        } 
        Trc.exit();
        return baseReader; 
    } 

	/**
	 * Get a reader for an imported wsdl document. Returns null if the import document
	 * cannot be located.
	 * @param base The document base uri for the parent wsdl document
	 * @param relativeLocation The relative uri of the import wsdl document
	 * @return The reader or null if the import cannot be resolved
	 */
    public Reader getImportReader(String base, String relativeLocation) {
    	Trc.entry(this,base,relativeLocation);
    	
    	// Reset importReader if finding import within import
    	importReader = null;
    	boolean triedSU = false; 
        try { 
            // If a ClassLoader was used to load the base document, chances 
            // are we need to use it to find the import. 
            URL url = null; 
            if (loader != null) { 
                if (relativeLocation.startsWith("/") || relativeLocation.startsWith("\\")) { 
                    // Relative location has been specified from a root dir. However, 
                    // using a ClassLoader, root dirs don't mean anything. 
                    relativeLocation = relativeLocation.substring(1, relativeLocation.length()); 
                    InputStream in = loader.getResourceAsStream(relativeLocation); 
                    importReader = new InputStreamReader(in); 
                } else if (relativeLocation.indexOf("://") != -1) { 
                    // This is a fully specified URL of some kind so don't use the 
                    // ClassLoader to find the import.
                    triedSU = true; 
                    url = StringUtils.getURL(null, relativeLocation); 
                    importReader = StringUtils.getContentAsReader(url); 
                } else { 
                    // Import location has been specified relative to the base document 
                    // and so we can to try to form the complete path to it. 						 
                    if (base != null) { 
                        int i = base.lastIndexOf("/"); 
                        if (i == -1) 
                            i = base.lastIndexOf("\\"); 
                        if (i != -1) { 
                            String path = base.substring(0, i + 1);
                            String resolvedPath = path + relativeLocation;
                            if (relativeLocation.startsWith("..")) {
                            	resolvedPath = resolvePath(path, relativeLocation);
                            }
                            if (resolvedPath == null) {
                            	throw new Exception("Invalid Path");
                            }
                            
                            // Make sure that resolved path starts with file:
                            if (resolvedPath.startsWith("file:")) { 
	                            url = new URL(null, resolvedPath);
                            } else {
                            	url = new URL(null, "file:" + resolvedPath);
                            }
                        } else { 
                            url = new URL(null, "file:" + relativeLocation); 
                        }  
                        InputStream in = loader.getResourceAsStream(url.getPath()); 
                        importReader = new InputStreamReader(in); 
                    } else {                    	
                    	url = new URL(null, "file:" + relativeLocation);
                    	InputStream in = loader.getResourceAsStream(url.getPath()); 
                        importReader = new InputStreamReader(in); 
                    }
                } 
            } else {
            	triedSU = true; 
                URL contextURL = (base != null) ? StringUtils.getURL(null, base) : null; 
                url = StringUtils.getURL(contextURL, relativeLocation); 
                importReader = StringUtils.getContentAsReader(url); 
            } 
            importBase = (url == null) ? relativeLocation : url.toString(); 
        } catch (Exception e) {
        	Trc.exception(e);
        	// If we have not tried using a non-ClassLoader route, try it now
        	// as a last resort.
        	if (!triedSU) {
        		try {        		
               	    URL contextURL = (base != null) ? StringUtils.getURL(null, base) : null; 
                    URL url = StringUtils.getURL(contextURL, relativeLocation); 
                    importReader = StringUtils.getContentAsReader(url);
                    importBase = (url == null) ? relativeLocation : url.toString();         		
        		} catch (Exception e2) {
        			Trc.exception(e2);
        			// we can't find the import so set a temporary value for the import URI. This is
        			// necessary to avoid a NullPointerException in WSDLReaderImpl
        			importBase = "unknownImportURI";
        		}
        	} else {
        	    // we can't find the import so set a temporary value for the import URI. This is
        	    // necessary to avoid a NullPointerException in WSDLReaderImpl
        		importBase = "unknownImportURI";        		
        	}
        } 
        Trc.exit();
        return importReader; 
    } 

	/**
	 * Get the document base uri for the base wsdl document
	 * @return The document base uri
	 */ 
    public String getBaseURI() { 
    	Trc.entry(this);
    	Trc.exit(documentBase);
        return documentBase; 
    } 

	/**
	 * Get the document base uri for the last import document to be resolved
	 * by this locator. This is useful if resolving imports within imports.
	 * @return The document base uri
	 */  
    public String getLatestImportURI() { 
    	Trc.entry(this);
    	Trc.exit(importBase);
        return importBase; 
    }

	/**
	 * Resolve a path when the relative location begins with ..
	 */     
    private String resolvePath(String ba, String rel) {
		StringBuffer sb = new StringBuffer(rel);
		int dd = 0;
		while(sb.length() > 0) {
			if(sb.length() > 3 && sb.charAt(0) == '.' && sb.charAt(1) == '.'
			&& (sb.charAt(2) == '/' || sb.charAt(2) == '\\')) {
				dd++;
				sb.delete(0,3);
			} else {
				break;
			}
		}
		StringBuffer sb2 = new StringBuffer(ba);
		int j = sb2.length()-1;
		int found = 0;
		for (int k = j; k>=0; k--) {
			if (k!=j && (sb2.charAt(k) == '/' || sb2.charAt(k) == '\\')) {
				found++;
			}
			if (found < dd) {
				sb2.deleteCharAt(k);
			} else {
				break;
			}
		}
		if (found+1 < dd) return null;
		return sb2.toString() + sb.toString();	
    }

    /**
     * Close any Reader objects that have been created
     * @throws IOException If a call to close() on one of the Reader objects fails
     */
    public void close() throws IOException {
   		if (baseReader != null) baseReader.close();
  		if (importReader != null) importReader.close();
    }    
}


